/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2007 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.01 of the PHP license,      |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_01.txt                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author:                                                              |
  +----------------------------------------------------------------------+
*/

/* $Id: header,v 1.16.2.1.2.1 2007/01/01 19:32:09 iliaa Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "php_cego.h"

#include "cgwrap.h"

#define PHPCEGOVERSION "1.0.1"

struct php_cego_db 
{
    CGDB *cgdb;
    int last_err_code;
    zend_bool is_persistent;
    long rsrc_id;
};

struct php_cego_stmt
{
    CGStmt *stmt;
    int last_err_code;
    zend_bool is_persistent;
    long rsrc_id;
};

struct php_cego_fetch
{
    CGFetch *fetch;
    int last_err_code;
    zend_bool is_persistent;
    long rsrc_id;
};

#define FLDBUFSIZE 2000

static ZEND_RSRC_DTOR_FUNC(php_cego_db_dtor)
{
    if (rsrc->ptr) 
    {
	struct php_cego_db *db = (struct php_cego_db*)rsrc->ptr;
	
	if ( db->cgdb)
	{
	    cego_disconnect(db->cgdb);
	    free(db->cgdb);
	}
	free(db);
	rsrc->ptr = NULL;
    }
}

static ZEND_RSRC_DTOR_FUNC(php_cego_stmt_dtor)
{
    if (rsrc->ptr) 
    {
	struct php_cego_stmt *cgstmt = (struct php_cego_stmt*)rsrc->ptr;
	
	if ( cgstmt->stmt)
	{
	    cego_free_stmt(cgstmt->stmt);
	}
	free(cgstmt);
	rsrc->ptr = NULL;
    }
}

static ZEND_RSRC_DTOR_FUNC(php_cego_fetch_dtor)
{
    if (rsrc->ptr) 
    {
	struct php_cego_fetch *cgfetch = (struct php_cego_fetch*)rsrc->ptr;

	if ( cgfetch->fetch)
	{
	    cego_abort(cgfetch->fetch);
	}
	free(cgfetch);
	rsrc->ptr = NULL;
    }
}


/* True global resources - no need for thread safety here */
static int le_cego, le_pcego;
static int le_cgstmt, le_pcgstmt;
static int le_cgfetch, le_pcgfetch;


#define DB_FROM_ZVAL(db, zv)  ZEND_FETCH_RESOURCE2(db, struct php_cego_db*, zv, -1, "cego database", le_cego, le_pcego)
#define STMT_FROM_ZVAL(stmt, zv)  ZEND_FETCH_RESOURCE2(stmt, struct php_cego_stmt*, zv, -1, "cego stmt", le_cgstmt, le_pcgstmt)
#define FETCH_FROM_ZVAL(fetch, zv)  ZEND_FETCH_RESOURCE2(fetch, struct php_cego_fetch*, zv, -1, "cego fetch", le_cgfetch, le_pcgfetch)


/* {{{ cego_functions[]
 *
 * Every user visible function must have an entry in cego_functions[].
 */
zend_function_entry cego_functions[] = {
    PHP_FE(confirm_cego_compiled,	NULL)		/* For testing, remove later. */
    PHP_FE(cego_phpcegoversion, NULL)
    PHP_FE(cego_connect, NULL)
    PHP_FE(cego_prepare, NULL)
    PHP_FE(cego_bind, NULL)
    PHP_FE(cego_execute, NULL)
    PHP_FE(cego_update, NULL)
    PHP_FE(cego_query, NULL)
    PHP_FE(cego_fetch, NULL)
    PHP_FE(cego_fieldinfo, NULL)
    PHP_FE(cego_release, NULL)
    PHP_FE(cego_close, NULL)
    PHP_FE(cego_error, NULL)
    {NULL, NULL, NULL}
};
/* }}} */

/* {{{ cego_module_entry
 */
zend_module_entry cego_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
    STANDARD_MODULE_HEADER,
#endif
    "cego",
    cego_functions,
    PHP_MINIT(cego),
    PHP_MSHUTDOWN(cego),
    PHP_RINIT(cego),		/* Replace with NULL if there's nothing to do at request start */
    PHP_RSHUTDOWN(cego),	/* Replace with NULL if there's nothing to do at request end */
    PHP_MINFO(cego),
#if ZEND_MODULE_API_NO >= 20010901
    "0.1", /* Replace with version number for your extension */
#endif
    STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_CEGO
ZEND_GET_MODULE(cego)
#endif

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(cego)
{
    /* If you have INI entries, uncomment these lines 
       REGISTER_INI_ENTRIES(); */
    
    
    le_cego = zend_register_list_destructors_ex(php_cego_db_dtor, NULL, "cego database", module_number);
    le_pcego= zend_register_list_destructors_ex(NULL, php_cego_db_dtor, "cego database (persistent)", module_number);
    
    le_cgstmt = zend_register_list_destructors_ex(php_cego_stmt_dtor, NULL, "cego stmt", module_number);
    le_pcgstmt= zend_register_list_destructors_ex(NULL, php_cego_stmt_dtor, "cego stmt (persistent)", module_number);
    
    le_cgfetch = zend_register_list_destructors_ex(php_cego_fetch_dtor, NULL, "cego fetch", module_number);
    le_pcgfetch= zend_register_list_destructors_ex(NULL, php_cego_fetch_dtor, "cego fetch (persistent)", module_number);


    REGISTER_LONG_CONSTANT("CEGO_INT",CG_INT, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("CEGO_VARCHAR",CG_VARCHAR, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("CEGO_LONG",CG_LONG, CONST_CS | CONST_PERSISTENT);
    
    
    return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 */
PHP_MSHUTDOWN_FUNCTION(cego)
{
    /* uncomment this line if you have INI entries
       UNREGISTER_INI_ENTRIES();
    */
    return SUCCESS;
}
/* }}} */

/* Remove if there's nothing to do at request start */
/* {{{ PHP_RINIT_FUNCTION
 */
PHP_RINIT_FUNCTION(cego)
{
    return SUCCESS;
}
/* }}} */

/* Remove if there's nothing to do at request end */
/* {{{ PHP_RSHUTDOWN_FUNCTION
 */
PHP_RSHUTDOWN_FUNCTION(cego)
{
    return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(cego)
{
    php_info_print_table_start();
    php_info_print_table_header(2, "cego support", "enabled");
    php_info_print_table_end();
    
    /* Remove comments if you have entries in php.ini
       DISPLAY_INI_ENTRIES();
    */
}
/* }}} */


/* Remove the following function when you have succesfully modified config.m4
   so that your module can be compiled into PHP, it exists only for testing
   purposes. */

/* Every user-visible function in PHP should document itself in the source */
/* {{{ proto string confirm_cego_compiled(string arg)
   Return a string to confirm that the module is compiled in */
PHP_FUNCTION(confirm_cego_compiled)
{
    char *arg = NULL;
    int arg_len, len;
    char *strg;
    
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &arg, &arg_len) == FAILURE) {
	return;
    }
    
    len = spprintf(&strg, 0, "Congratulations! You have successfully modified ext/%.78s/config.m4. Module %.78s is now compiled into PHP.", "cego", arg);
    RETURN_STRINGL(strg, len, 0);
}

PHP_FUNCTION(cego_phpcegoversion)
{
    RETURN_STRING(PHPCEGOVERSION, 1);
}

/* proto resource cego_connect(string hostname, int portno, string tableset, string user, string passwd)

   hostname - database host to connect to
   portno - port number to connect to
   tableset - target tableset
   user - database user
   passwd - database user password

   -->> if successful, returns a valid database connection 
*/

PHP_FUNCTION(cego_connect)
{
    
    char *host=NULL;
    char *tableset=NULL;
    char *user=NULL;
    char *passwd=NULL;
    int host_len;
    int tableset_len;
    int user_len;
    int passwd_len;
    
    long portno = 0;
    
    struct php_cego_db *db = NULL;
    
    char *logfile=NULL; // "cgwrap.log";
    
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "slsss", 
			      &host, &host_len,
			      &portno,
			      &tableset, &tableset_len,
			      &user, &user_len,
			      &passwd, &passwd_len)==FAILURE) {
	php_std_error_handling();
	return;
    }
    
    CGDB *cgdb;
    
    php_error_docref(NULL TSRMLS_CC, E_NOTICE, "Connecting to host");    
    
    if ( (cgdb = cego_connect(host, portno, tableset, user, passwd, logfile)) == NULL )
    {		
		RETURN_FALSE;
    }
    
    // cego_debug_level(cgdb);
    
    db = (struct php_cego_db *)malloc(sizeof(struct php_cego_db));
    db->is_persistent =  0;
    db->last_err_code = 0;
    db->cgdb = cgdb;
    
    php_error_docref(NULL TSRMLS_CC, E_NOTICE, "Connect ok");    
    
    
    ZEND_REGISTER_RESOURCE(return_value, db, le_cego);
    
}


/* proto resource cego_prepare(string stmt)
   
   stmt - SQL statement to prepare

   -->> if successful, returns a valid statement resource

*/

PHP_FUNCTION(cego_prepare)
{
    char *stmt;
    int stmt_len;
    
    struct php_cego_stmt *cgstmt = NULL;
    
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &stmt, &stmt_len) == FAILURE ) 
    {
	php_std_error_handling();
	return;
    }
    
    CGStmt *pStmt;
    pStmt = cego_prepare(stmt);
    
    cgstmt = (struct php_cego_stmt *)malloc(sizeof(struct php_cego_stmt));
    cgstmt->is_persistent =  0;
    cgstmt->last_err_code = 0;
    cgstmt->stmt = pStmt;
    
    ZEND_REGISTER_RESOURCE(return_value, cgstmt, le_cgstmt);	
    
}

/* proto bool cego_bind(resource stmt, string var, int type, int pos)
   
   stmt - statement resource to bind
   var - bind value
   type - cego datatype
   pos - bind position in statement

   -->> if successful, true is returned, else false

*/

PHP_FUNCTION(cego_bind)
{
    zval *zstmt;
    char *val;
    int val_len;
    long type;
    long pos;
    
    struct php_cego_stmt *cgstmt = NULL;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rsll", 
			      &zstmt, &val, &val_len, &type, &pos) == FAILURE ) 
    {
	php_std_error_handling();
	RETURN_FALSE;
    }

    STMT_FROM_ZVAL(cgstmt, &zstmt);

    CGVal* pCGval = (CGVal*)malloc(sizeof(CGVal));

    if ( type == CG_INT )
    {
	int *pI = (int*)malloc(sizeof(int));
	*pI = atoi(val);

	pCGval->type = CG_INT;
	pCGval->len = sizeof(int);
	pCGval->val = pI;
    }
    else if ( type == CG_VARCHAR )
    {	
	char *pS = (char*)malloc(val_len);
	strcpy(pS, val);
	pCGval->type = CG_VARCHAR;
	pCGval->len = val_len;
	pCGval->val = pS;
	
    }
    else // TODO : add more type here
    {	
	pCGval->type = CG_NULL;
	pCGval->len = 0;
	pCGval->val = NULL;
    }
    
    cego_bind_in(cgstmt->stmt, pCGval, pos); 

    RETURN_TRUE;
    
}


/* proto resource cego_execute(resource db, resource stmt)

   db - database resource
   stmt - stmt resource
   
   -->> if successful, returns a valid fetch resource 
   
*/

PHP_FUNCTION(cego_execute)
{
    zval *zdb;
    zval *zstmt;
    struct php_cego_db *db;
    struct php_cego_stmt *cgstmt;
    struct php_cego_fetch *cgfetch;
    
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rr", &zdb, &zstmt) == FAILURE ) 
    {
	php_std_error_handling();
	return;
    }
    
    DB_FROM_ZVAL(db, &zdb);
    STMT_FROM_ZVAL(cgstmt, &zstmt);
    
    
    CGFetch* pFetch = cego_allocate_fetch();
    
    if ( cego_execute(db->cgdb, cgstmt->stmt, pFetch) != 0 )
    {
	cego_free_fetch(pFetch);
	RETURN_FALSE;
    }
    
    cgfetch = (struct php_cego_fetch *)malloc(sizeof(struct php_cego_fetch));
    cgfetch->is_persistent =  0;
    cgfetch->last_err_code = 0;
    cgfetch->fetch = pFetch;
    
    ZEND_REGISTER_RESOURCE(return_value, cgfetch, le_cgfetch);	
    
}


/* proto resource cego_update(resource db, resource stmt)

   db - database resource
   stmt - stmt resource
   
   -->> if successful, returns a true, else false
   
*/

PHP_FUNCTION(cego_update)
{
    zval *zdb;
    zval *zstmt;
    struct php_cego_db *db;
    struct php_cego_stmt *cgstmt;
    
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rr", &zdb, &zstmt) == FAILURE ) 
    {
	php_std_error_handling();
	return;
    }
    
    DB_FROM_ZVAL(db, &zdb);
    STMT_FROM_ZVAL(cgstmt, &zstmt);
    
       
    if ( cego_execute(db->cgdb, cgstmt->stmt, NULL) != 0 )
    {
	RETURN_FALSE;
    }
    
    RETURN_TRUE;
    
}


/* proto array cego_query(resource db, string query)

   db - database resource
   query - query stmt string

   -->> if successful, returns true, else false 

*/

PHP_FUNCTION(cego_query)
{
    zval *zdb;
    struct php_cego_db *db;

    char *query;
    int query_len;
    
    struct php_cego_stmt *cgstmt = NULL;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", 
			      &zdb, &query, &query_len) == FAILURE ) 
    {
	php_std_error_handling();
	RETURN_FALSE;
    }
        
    DB_FROM_ZVAL(db, &zdb);
    
    if ( cego_query(db->cgdb, query, 0) != 0 )
    {
	RETURN_FALSE;
    }

    RETURN_TRUE;

}


/* proto array cego_fetch(resource fetch)

   fetch - fetch resource

   -->> if successful, returns a value array 

*/


PHP_FUNCTION(cego_fetch)
{
    
    zval *zfetch;
    struct php_cego_fetch *cgfetch;

    char tmpBuf[FLDBUFSIZE];
    int i;
    
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zfetch) == FAILURE ) 
    {
	php_std_error_handling();
	return;
    }
    
    FETCH_FROM_ZVAL(cgfetch, &zfetch);

    int ncol = cego_num_col(cgfetch->fetch);
    
    CGVal* valArray = malloc(ncol * sizeof(CGVal));
    
    for ( i=0;i<ncol;i++)
	valArray[i].val=0;
    
    int colnum;
    colnum = cego_fetch (cgfetch->fetch, valArray, ncol);
    
    if ( colnum == 0 )
    {
	cego_free_fetch(cgfetch->fetch);
	cgfetch->fetch = NULL;	
	free(valArray);
	RETURN_FALSE;
    }
    
    array_init(return_value);
    
    i=0;
    while ( i < colnum )
    {
	zval *fvalue;
	MAKE_STD_ZVAL(fvalue);
	
	if (valArray[i].type == CG_NULL) 
	{
	    ZVAL_NULL(fvalue);
	} 
	else if (valArray[i].type == CG_INT) 
	{
	    
	    int ival = *(int*)valArray[i].val;			
	    int len = sprintf(tmpBuf, "%d", ival);
	    
	    Z_STRVAL_P(fvalue) = emalloc(len);
	    strcpy(Z_STRVAL_P(fvalue), tmpBuf);
	    Z_STRLEN_P(fvalue) = len;
	    Z_TYPE_P(fvalue) = IS_STRING;
	    
	} 
	else if (valArray[i].type == CG_VARCHAR)
	{
	    int len = sprintf(tmpBuf, "%s", valArray[i].val);
	    
	    Z_STRVAL_P(fvalue) = emalloc(len);
	    strcpy(Z_STRVAL_P(fvalue), tmpBuf);
	    Z_STRLEN_P(fvalue) = len;
	    Z_TYPE_P(fvalue) = IS_STRING;
	}
	else // TODO : more datatypes
	{
	    ZVAL_NULL(fvalue);
	}
	
	char* colName = (char*)cego_getcol(cgfetch->fetch,i);
	add_index_zval(return_value, i, fvalue);
	ZVAL_ADDREF(fvalue);
	add_assoc_zval(return_value, colName, fvalue);
	
	// add_next_index_zval(return_value, fvalue);
	
	i++;
    }
    
    free(valArray);
}


/* proto array cego_fieldinfo(resource fetch)

   fetch - fetch resource

   -->> if successful, returns a field array 

*/


PHP_FUNCTION(cego_fieldinfo)
{
    
    zval *zfetch;
    struct php_cego_fetch *cgfetch;

    int i;
    
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zfetch) == FAILURE ) 
    {
	php_std_error_handling();
	return;
    }
    
    FETCH_FROM_ZVAL(cgfetch, &zfetch);
    
    int ncol = cego_num_col(cgfetch->fetch);
        
    array_init(return_value);
    
    i=0;
    while ( i < ncol)
    {
	zval *fvalue;
	MAKE_STD_ZVAL(fvalue);

	char* colName = (char*)cego_getcol(cgfetch->fetch,i);	
	int len = strlen(colName);
	Z_STRVAL_P(fvalue) = emalloc(len);
	strcpy(Z_STRVAL_P(fvalue), colName);
	Z_STRLEN_P(fvalue) = len;
	Z_TYPE_P(fvalue) = IS_STRING;
	
	i++;

	add_index_zval(return_value, i, fvalue);
    }    
}



/* proto bool cego_release(resource fetch)

   fetch - fetch resource

   -->> if successful, returns true, else false

*/

PHP_FUNCTION(cego_release)
{

    zval *zfetch;
    struct php_cego_fetch *cgfetch;
    
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zfetch) == FAILURE ) 
    {
	php_std_error_handling();
	RETURN_FALSE;
    }
    
    FETCH_FROM_ZVAL(cgfetch, &zfetch);

    if ( cgfetch->fetch == NULL)
	return;

    cego_abort(cgfetch->fetch);

    RETURN_TRUE;
}



/* proto bool cego_close(resource db)

   db - database resource

   -->> if successful, returns true, else false

*/

PHP_FUNCTION(cego_close)
{
    
    zval *zdb;
    struct php_cego_db *db;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zdb) == FAILURE ) 
    {
	php_std_error_handling();
	RETURN_TRUE;
    }
    DB_FROM_ZVAL(db, &zdb);
    
    cego_disconnect(db->cgdb);
    
    free(db->cgdb);
    db->cgdb = NULL;
    
    zend_list_delete(Z_RESVAL_P(zdb));

    RETURN_FALSE;    

}

/* proto string cego_error()


   -->> returns last error message

*/

PHP_FUNCTION(cego_error)
{
    RETURN_STRING((char*)cego_error(), 1);
}
